@@ -16,6 +16,20 @@
  *	Port 9100+n will then be passively opened
  *	n defaults to 0
  *
+ *	Version 0.97
+ *	Patches by Stefan Sichler.
+ *	Stream to printer is only closed after EOF from network if 
+ *	it is no more busy, otherwise printer driver may discard data of last 
+ *	write() at close().
+ *
+ *	Version 0.96
+ *	Patches by Stefan Sichler.
+ *	Fixed bi-directional mode to stay alive until connection is closed.
+ *	Fixed timeout value in select() (controlling in/out balancing).
+ *	Fixed network-to-printer buffer reporting EOF although there was data 
+ *	left in the buffer.
+ *	Added log-to-stdout option (-d).
+ *
  *	Version 0.95
  *	Patch by Mario Izquierdo
  *	Fix incomplete conversion to manipulate new ip_addr structure
@@ -103,6 +117,7 @@
 #include	<netdb.h>
 #include	<syslog.h>
 #include	<errno.h>
+#include	<stdarg.h>
 #include	<sys/types.h>
 #include	<sys/time.h>
 #include	<sys/resource.h>
@@ -141,18 +156,20 @@ typedef struct {
 	int bytes;		/* The number of bytes currently buffered. */
 	int totalin;		/* Total bytes that have been read. */
 	int totalout;		/* Total bytes that have been written. */
-	int eof;		/* Nonzero indicates the input file has reached EOF. */
+	int eof_read;		/* Nonzero indicates the input file has reached EOF. */
+	int eof_sent;		/* Nonzero indicates the output file has fully received all data. */
 	int err;		/* Nonzero indicates an error detected on the output file. */
 	char buffer[BUFFER_SIZE];	/* Buffered data goes here. */
 } Buffer_t;
 
 static char *progname;
-static char version[] = "Version 0.94";
-static char copyright[] = "Copyright (c) 2008 Ken Yap, GPLv2";
+static char version[] = "Version 0.97";
+static char copyright[] = "Copyright (c) 2008-2014 Ken Yap and others, GPLv2";
 static int lockfd = -1;
 static char *device = 0;
 static int bidir = 0;
 static char *bindaddr = 0;
+static int log_to_stdout = 0;
 
 
 /* Helper function: convert a struct sockaddr address (IPv4 and IPv6) to a string */
@@ -191,7 +208,7 @@ uint16_t get_port(const struct sockaddr *sa)
 void usage(void)
 {
 	fprintf(stderr, "%s %s %s\n", progname, version, copyright);
-	fprintf(stderr, "Usage: %s [-f device] [-i bindaddr] [-bv] [0|1|2]\n", progname);
+	fprintf(stderr, "Usage: %s [-f device] [-i bindaddr] [-bvd] [0|1|2]\n", progname);
 	exit(1);
 }
 
@@ -200,6 +217,17 @@ void show_version(void)
 	fprintf(stdout, "%s %s\n", progname, version);
 }
 
+void dolog(int level, char* msg, ...)
+{
+	va_list argp;
+	va_start(argp, msg);
+	if (log_to_stdout)
+		vfprintf(stdout, msg, argp);
+	else if (level != LOG_DEBUG)
+		vsyslog(level, msg, argp);
+	va_end(argp);	
+}
+
 int open_printer(int lpnumber)
 {
 	int lp;
@@ -212,10 +240,10 @@ int open_printer(int lpnumber)
 #endif
 	if (device == 0)
 		device = lpname;
-	if ((lp = open(device, bidir ? O_RDWR : O_WRONLY)) == -1) {
+	if ((lp = open(device, bidir ? (O_RDWR|O_NONBLOCK) : O_WRONLY)) == -1) {
 		if (errno != EBUSY)
-			syslog(LOGOPTS, "%s: %m\n", device);
-		syslog(LOGOPTS, "%s: %m, will try opening later\n", device);
+			dolog(LOGOPTS, "%s: %m\n", device);
+		dolog(LOGOPTS, "%s: %m, will try opening later\n", device);
 	}
 	return (lp);
 }
@@ -227,14 +255,14 @@ int get_lock(int lpnumber)
 
 	(void)snprintf(lockname, sizeof(lockname), LOCKFILE, lpnumber);
 	if ((lockfd = open(lockname, O_CREAT | O_RDWR, 0666)) < 0) {
-		syslog(LOGOPTS, "%s: %m\n", lockname);
+		dolog(LOGOPTS, "%s: %m\n", lockname);
 		return (0);
 	}
 	memset(&lplock, 0, sizeof(lplock));
 	lplock.l_type = F_WRLCK;
 	lplock.l_pid = getpid();
 	if (fcntl(lockfd, F_SETLKW, &lplock) < 0) {
-		syslog(LOGOPTS, "%s: %m\n", lockname);
+		dolog(LOGOPTS, "%s: %m\n", lockname);
 		return (0);
 	}
 	return (1);
@@ -257,17 +285,18 @@ void initBuffer(Buffer_t * b, int infd, int outfd, int detectEof)
 	b->bytes = 0;
 	b->totalin = 0;
 	b->totalout = 0;
-	b->eof = 0;
+	b->eof_read = 0;
+	b->eof_sent = 0;
 	b->err = 0;
 }
 
 /* Sets the readfds and writefds (used by select) based on current buffer state. */
 void prepBuffer(Buffer_t * b, fd_set * readfds, fd_set * writefds)
 {
-	if (!b->err && b->bytes != 0) {
+	if (b->outfd>=0 && !b->err && (b->bytes != 0 || b->eof_read)) {
 		FD_SET(b->outfd, writefds);
 	}
-	if (!b->eof && b->bytes < sizeof(b->buffer)) {
+	if (b->infd>=0 && !b->eof_read && b->bytes < sizeof(b->buffer)) {
 		FD_SET(b->infd, readfds);
 	}
 }
@@ -276,7 +305,7 @@ void prepBuffer(Buffer_t * b, fd_set * readfds, fd_set * writefds)
 ssize_t readBuffer(Buffer_t * b)
 {
 	int avail;
-	ssize_t result = 1;
+	ssize_t result = 0;
 	/* If err, the data will not be written, so no need to store it. */
 	if (b->bytes == 0 || b->err) {
 		/* The buffer is empty. */
@@ -303,20 +332,25 @@ ssize_t readBuffer(Buffer_t * b)
 				/* Time to wrap the buffer. */
 				b->endidx = 0;
 			}
-		} else if (result < 0 || b->detectEof) {
-			/* Mark end-of-file on input error, or on zero return if detectEof. */
-			b->eof = 1;
+		} else if (result < 0) {
+			dolog(LOGOPTS, "read: %m\n");
+			b->err = 1;
 		}
+		else if (b->detectEof) {
+			dolog(LOG_DEBUG, "read: eof\n");
+			b->eof_read = 1;
+		} else
+			result = 0; // in case there is still data in the buffer, ignore the error by now
 	}
-	/* Return the value returned by read(), which is -1, 0, or #bytes read. */
+	/* Return the value returned by read(), which is -1 (error), or #bytes read. */
 	return result;
 }
 
-/* Writes data from a buffer to the output file. */
+/* Writes data from a buffer to the output file or discard if no output file is set. */
 ssize_t writeBuffer(Buffer_t * b)
 {
 	int avail;
-	ssize_t result = 1;
+	ssize_t result = 0;
 	if (b->bytes == 0 || b->err) {
 		/* Buffer is empty. */
 		avail = 0;
@@ -328,15 +362,19 @@ ssize_t writeBuffer(Buffer_t * b)
 		avail = sizeof(b->buffer) - b->startidx;
 	}
 	if (avail) {
-		result = write(b->outfd, b->buffer + b->startidx, avail);
+		if (b->outfd>=0)
+			result = write(b->outfd, b->buffer + b->startidx, avail);
+		else
+			result = avail;
 		if (result < 0) {
 			/* Mark the output file in an error condition. */
-			syslog(LOGOPTS, "write: %m\n");
+			dolog(LOGOPTS, "write: %m\n");
 			b->err = 1;
 		} else {
 			/* Zero or more bytes were written. */
 			b->startidx += result;
-			b->totalout += result;
+			if (b->outfd>=0)
+				b->totalout += result;
 			b->bytes -= result;
 			if (b->startidx == sizeof(b->buffer)) {
 				/* Unwrap the buffer. */
@@ -344,7 +382,10 @@ ssize_t writeBuffer(Buffer_t * b)
 			}
 		}
 	}
-	/* Return the write() result, -1 (error), or #bytes written. */
+	else if (b->eof_read)
+		b->eof_sent = 1;
+	
+	/* Return the write() result, -1 (error) or #bytes written. */
 	return result;
 }
 
@@ -365,43 +406,41 @@ int copy_stream(int fd, int lp)
 		initBuffer(&printerToNetworkBuffer, lp, fd, 0);
 		fd_set readfds;
 		fd_set writefds;
-		/* Initially read from both streams, don't write to either. */
-		FD_ZERO(&readfds);
-		FD_ZERO(&writefds);
-		FD_SET(lp, &readfds);
-		FD_SET(fd, &readfds);
-		/* Finish when no longer reading fd, and no longer writing to lp. */
-		/* Although the printer to network stream may not be finished, that does not matter. */
-		while ((FD_ISSET(fd, &readfds)) || (FD_ISSET(lp, &writefds))) {
+		/* Finish when network sent EOF. */
+		/* Although the printer to network stream may not be finished (does this matter?) */
+		while (!networkToPrinterBuffer.eof_sent && !networkToPrinterBuffer.err) {
+			FD_ZERO(&readfds);
+			FD_ZERO(&writefds);
+			prepBuffer(&networkToPrinterBuffer, &readfds, &writefds);
+			prepBuffer(&printerToNetworkBuffer, &readfds, &writefds);
+
 			int maxfd = lp > fd ? lp : fd;
 			if (timer) {
 				/* Delay after reading from the printer, so the */
 				/* return stream cannot dominate. */
 				/* Don't read from the printer until the timer expires. */
 				gettimeofday(&now, 0);
-				if ((now.tv_sec > then.tv_sec) || (now.tv_sec == then.tv_sec && now.tv_usec > then.tv_usec)) {
+				if ((now.tv_sec > then.tv_sec) || (now.tv_sec == then.tv_sec && now.tv_usec > then.tv_usec))
 					timer = 0;
-				} else {
-					timeout.tv_sec = then.tv_sec;
-					timeout.tv_usec = then.tv_usec;
+				else
 					FD_CLR(lp, &readfds);
-				}
-			}
-			if (timer) {
-				result = select(maxfd + 1, &readfds, &writefds, 0, &timeout);
-			} else {
-				result = select(maxfd + 1, &readfds, &writefds, 0, 0);
 			}
+			timeout.tv_sec = 0;
+			timeout.tv_usec = 100000;
+			result = select(maxfd + 1, &readfds, &writefds, 0, &timeout);
 			if (result < 0)
 				return (result);
 			if (FD_ISSET(fd, &readfds)) {
 				/* Read network data. */
 				result = readBuffer(&networkToPrinterBuffer);
+				if (result > 0)
+					dolog(LOG_DEBUG,"read %d bytes from network\n",result);
 			}
 			if (FD_ISSET(lp, &readfds)) {
 				/* Read printer data, but pace it more slowly. */
 				result = readBuffer(&printerToNetworkBuffer);
-				if (result >= 0) {
+				if (result > 0) {
+					dolog(LOG_DEBUG,"read %d bytes from printer\n",result);
 					gettimeofday(&then, 0);
 					// wait 100 msec before reading again.
 					then.tv_usec += 100000;
@@ -415,32 +454,43 @@ int copy_stream(int fd, int lp)
 			if (FD_ISSET(lp, &writefds)) {
 				/* Write data to printer. */
 				result = writeBuffer(&networkToPrinterBuffer);
+				if (result > 0)
+					dolog(LOG_DEBUG,"wrote %d bytes to printer\n",result);
 			}
-			if (FD_ISSET(fd, &writefds)) {
+			if (FD_ISSET(fd, &writefds) || printerToNetworkBuffer.outfd == -1) {
 				/* Write data to network. */
 				result = writeBuffer(&printerToNetworkBuffer);
-				/* If socket write error, stop reading from printer */
-				if (result < 0)
-					networkToPrinterBuffer.eof = 1;
+				/* If socket write error, discard further data from printer */
+				if (result < 0) {
+					printerToNetworkBuffer.outfd = -1;
+					printerToNetworkBuffer.err = 0;
+					result = 0;
+					dolog(LOG_DEBUG,"network write error, discarding further printer data\n",result);
+				}
+				else if (result > 0) {
+					if (printerToNetworkBuffer.outfd == -1)
+						dolog(LOG_DEBUG,"discarded %d bytes from printer\n",result);				
+					else
+						dolog(LOG_DEBUG,"wrote %d bytes to network\n",result);
+				}
 			}
-			/* Prepare for next iteration. */
-			FD_ZERO(&readfds);
-			FD_ZERO(&writefds);
-			prepBuffer(&networkToPrinterBuffer, &readfds, &writefds);
-			prepBuffer(&printerToNetworkBuffer, &readfds, &writefds);
 		}
-		syslog(LOG_NOTICE,
-		       "Finished job: %d bytes received, %d bytes sent\n",
-		       networkToPrinterBuffer.totalout, printerToNetworkBuffer.totalout);
-		return (0);
+		dolog(LOG_NOTICE,
+		       "Finished job: %d/%d bytes sent to printer, %d/%d bytes sent to network\n",
+		       networkToPrinterBuffer.totalout,networkToPrinterBuffer.totalin, printerToNetworkBuffer.totalout, printerToNetworkBuffer.totalin);
 	} else {
 		/* Unidirectional: simply read from network, and write to printer. */
-		while ((result = readBuffer(&networkToPrinterBuffer)) > 0) {
-			(void)writeBuffer(&networkToPrinterBuffer);
+		while (!networkToPrinterBuffer.eof_sent && !networkToPrinterBuffer.err) {
+			result = readBuffer(&networkToPrinterBuffer);
+			if (result > 0)
+				dolog(LOG_DEBUG,"read %d bytes from network\n",result);
+			result = writeBuffer(&networkToPrinterBuffer);
+			if (result > 0)
+				dolog(LOG_DEBUG,"wrote %d bytes to printer\n",result);
 		}
-		syslog(LOG_NOTICE, "Finished job: %d bytes received\n", networkToPrinterBuffer.totalout);
-		return (result);
+		dolog(LOG_NOTICE, "Finished job: %d/%d bytes sent to printer\n", networkToPrinterBuffer.totalout, networkToPrinterBuffer.totalin);
 	}
+	return (networkToPrinterBuffer.err?-1:0);
 }
 
 void one_job(int lpnumber)
@@ -451,7 +501,7 @@ void one_job(int lpnumber)
 
 	if (getpeername(0, (struct sockaddr *)&client, &clientlen) >= 0) {
 		char host[INET6_ADDRSTRLEN];
-		syslog(LOG_NOTICE, "Connection from %s port %hu\n", get_ip_str((struct sockaddr *)&client, host, sizeof(host)), get_port((struct sockaddr *)&client));
+		dolog(LOG_NOTICE, "Connection from %s port %hu\n", get_ip_str((struct sockaddr *)&client, host, sizeof(host)), get_port((struct sockaddr *)&client));
 	}
 	if (get_lock(lpnumber) == 0)
 		return;
@@ -459,7 +509,7 @@ void one_job(int lpnumber)
 	while ((lp = open_printer(lpnumber)) == -1)
 		sleep(10);
 	if (copy_stream(0, lp) < 0)
-		syslog(LOGOPTS, "copy_stream: %m\n");
+		dolog(LOGOPTS, "copy_stream: %m\n");
 	close(lp);
 	free_lock();
 }
@@ -480,39 +530,42 @@ void server(int lpnumber)
 	const int bufsiz = 65536;
 
 #ifndef	TESTING
-	switch (fork()) {
-	case -1:
-		syslog(LOGOPTS, "fork: %m\n");
-		exit(1);
-	case 0:		/* child */
-		break;
-	default:		/* parent */
-		exit(0);
+	if (!log_to_stdout)
+	{
+		switch (fork()) {
+		case -1:
+			dolog(LOGOPTS, "fork: %m\n");
+			exit(1);
+		case 0:		/* child */
+			break;
+		default:		/* parent */
+			exit(0);
+		}
+		/* Now in child process */
+		resourcelimit.rlim_max = 0;
+		if (getrlimit(RLIMIT_NOFILE, &resourcelimit) < 0) {
+			dolog(LOGOPTS, "getrlimit: %m\n");
+			exit(1);
+		}
+		for (fd = 0; fd < resourcelimit.rlim_max; ++fd)
+			(void)close(fd);
+		if (setsid() < 0) {
+			dolog(LOGOPTS, "setsid: %m\n");
+			exit(1);
+		}
+		(void)chdir("/");
+		(void)umask(022);
+		fd = open("/dev/null", O_RDWR);	/* stdin */
+		(void)dup(fd);		/* stdout */
+		(void)dup(fd);		/* stderr */
+		(void)snprintf(pidfilename, sizeof(pidfilename), PIDFILE, lpnumber);
+		if ((f = fopen(pidfilename, "w")) == NULL) {
+			dolog(LOGOPTS, "%s: %m\n", pidfilename);
+			exit(1);
+		}
+		(void)fprintf(f, "%d\n", getpid());
+		(void)fclose(f);
 	}
-	/* Now in child process */
-	resourcelimit.rlim_max = 0;
-	if (getrlimit(RLIMIT_NOFILE, &resourcelimit) < 0) {
-		syslog(LOGOPTS, "getrlimit: %m\n");
-		exit(1);
-	}
-	for (fd = 0; fd < resourcelimit.rlim_max; ++fd)
-		(void)close(fd);
-	if (setsid() < 0) {
-		syslog(LOGOPTS, "setsid: %m\n");
-		exit(1);
-	}
-	(void)chdir("/");
-	(void)umask(022);
-	fd = open("/dev/null", O_RDWR);	/* stdin */
-	(void)dup(fd);		/* stdout */
-	(void)dup(fd);		/* stderr */
-	(void)snprintf(pidfilename, sizeof(pidfilename), PIDFILE, lpnumber);
-	if ((f = fopen(pidfilename, "w")) == NULL) {
-		syslog(LOGOPTS, "%s: %m\n", pidfilename);
-		exit(1);
-	}
-	(void)fprintf(f, "%d\n", getpid());
-	(void)fclose(f);
 	if (get_lock(lpnumber) == 0)
 		exit(1);
 #endif
@@ -522,7 +575,7 @@ void server(int lpnumber)
 	hints.ai_socktype = SOCK_STREAM;
 	(void)snprintf(service, sizeof(service), "%hu", (BASEPORT + lpnumber - '0'));
 	if (getaddrinfo(bindaddr, service, &hints, &res) != 0) {
-		syslog(LOGOPTS, "getaddr: %m\n");
+		dolog(LOGOPTS, "getaddr: %m\n");
 		exit(1);
 	}
 	ressave = res;
@@ -530,7 +583,7 @@ void server(int lpnumber)
 #ifdef	USE_GETPROTOBYNAME
 		if ((proto = getprotobyname("tcp6")) == NULL) {
 			if ((proto = getprotobyname("tcp")) == NULL) {
-				syslog(LOGOPTS, "Cannot find protocol for TCP!\n");
+				dolog(LOGOPTS, "Cannot find protocol for TCP!\n");
 				exit(1);
 			}
 		}
@@ -538,34 +591,34 @@ void server(int lpnumber)
 #else
 		if ((netfd = socket(res->ai_family, res->ai_socktype, IPPROTO_IP)) < 0)
 #endif
-	{
-		syslog(LOGOPTS, "socket: %m\n");
+		{
+			dolog(LOGOPTS, "socket: %m\n");
 			close(netfd);
 			res = res->ai_next;
 			continue;
 		}
 		if (setsockopt(netfd, SOL_SOCKET, SO_RCVBUF, &bufsiz, sizeof(bufsiz)) < 0) {
-			syslog(LOGOPTS, "setsocketopt: SO_RCVBUF: %m\n");
+			dolog(LOGOPTS, "setsocketopt: SO_RCVBUF: %m\n");
 			/* not fatal if it fails */
 		}
 		if (setsockopt(netfd, SOL_SOCKET, SO_SNDBUF, &bufsiz, sizeof(bufsiz)) < 0) {
-			syslog(LOGOPTS, "setsocketopt: SO_SNDBUF: %m\n");
+			dolog(LOGOPTS, "setsocketopt: SO_SNDBUF: %m\n");
 			/* not fatal if it fails */
 		}
 		if (setsockopt(netfd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) < 0) {
-			syslog(LOGOPTS, "setsocketopt: SO_REUSEADDR: %m\n");
+			dolog(LOGOPTS, "setsocketopt: SO_REUSEADDR: %m\n");
 			close(netfd);
 			res = res->ai_next;
 			continue;
 		}
 		if (bind(netfd, res->ai_addr, res->ai_addrlen) < 0) {
-			syslog(LOGOPTS, "bind: %m\n");
+			dolog(LOGOPTS, "bind: %m\n");
 			close(netfd);
 			res = res->ai_next;
 			continue;
 		}
 		if (listen(netfd, 5) < 0) {
-			syslog(LOGOPTS, "listen: %m\n");
+			dolog(LOGOPTS, "listen: %m\n");
 			close(netfd);
 			res = res->ai_next;
 			continue;
@@ -579,13 +632,13 @@ void server(int lpnumber)
 		char host[INET6_ADDRSTRLEN];
 #ifdef	USE_LIBWRAP
 		if (hosts_ctl("p910nd", STRING_UNKNOWN, get_ip_str((struct sockaddr *)&client, host, sizeof(host)), STRING_UNKNOWN) == 0) {
-			syslog(LOGOPTS,
+			dolog(LOGOPTS,
 			       "Connection from %s port %hu rejected\n", get_ip_str((struct sockaddr *)&client, host, sizeof(host)), get_port((struct sockaddr *)&client));
 			close(fd);
 			continue;
 		}
 #endif
-		syslog(LOG_NOTICE, "Connection from %s port %hu accepted\n", get_ip_str((struct sockaddr *)&client, host, sizeof(host)), get_port((struct sockaddr *)&client));
+		dolog(LOG_NOTICE, "Connection from %s port %hu accepted\n", get_ip_str((struct sockaddr *)&client, host, sizeof(host)), get_port((struct sockaddr *)&client));
 		/*write(fd, "Printing", 8); */
 
 		/* Make sure lp device is open... */
@@ -593,11 +646,11 @@ void server(int lpnumber)
 			sleep(10);
 
 		if (copy_stream(fd, lp) < 0)
-			syslog(LOGOPTS, "copy_stream: %m\n");
+			dolog(LOGOPTS, "copy_stream: %m\n");
 		(void)close(fd);
 		(void)close(lp);
 	}
-	syslog(LOGOPTS, "accept: %m\n");
+	dolog(LOGOPTS, "accept: %m\n");
 	free_lock();
 	exit(1);
 }
@@ -618,7 +671,7 @@ int is_standalone(void)
 	if (getsockname(0, (struct sockaddr *)&bind_addr, &ba_len) == 0)
 		return (0);	/* under (x)inetd */
 	if (errno != ENOTSOCK)	/* strange... */
-		syslog(LOGOPTS, "getsockname: %m\n");
+		dolog(LOGOPTS, "getsockname: %m\n");
 	return (1);
 }
 
@@ -635,11 +688,14 @@ int main(int argc, char *argv[])
 			progname = p + 1;
 	}
 	lpnumber = '0';
-	while ((c = getopt(argc, argv, "bi:f:v")) != EOF) {
+	while ((c = getopt(argc, argv, "bdi:f:v")) != EOF) {
 		switch (c) {
 		case 'b':
 			bidir = 1;
 			break;
+		case 'd':
+			log_to_stdout = 1;
+			break;
 		case 'f':
 			device = optarg;
 			break;
@@ -668,8 +724,10 @@ int main(int argc, char *argv[])
 	 * syslog ignored the LOG_PID and LOG_PERROR option.  I.e. the intention
 	 * was to add both options but the effect was to have neither.
 	 * I disagree with the intention to add PERROR.  --Stef  */
-	openlog(p, LOG_PID, LOG_LPR);
-	if (is_standalone())
+	if (!log_to_stdout)
+		openlog(p, LOG_PID, LOG_LPR);
+
+	if (log_to_stdout || is_standalone())
 		server(lpnumber);
 	else
 		one_job(lpnumber);
